;;;; scexp - a simple sexp->C translator, possibly as the backend to a compiler
;;;; Copyright 2004 Brian Mastenbrook
;;;; For details, please see the file LICENSE

;;; scexp depends on the common-idioms package:
;;; http://www.cs.indiana.edu/~bmastenb/software/common-idioms/

(defpackage :scexp (:use :common-lisp :common-idioms)
            (:export :translate-file :install-scexp-read-macro))
(defpackage :scexp-user (:use :common-lisp :scexp))
(defpackage :scexp-read (:use))
(in-package :scexp)

(defun quote-reader (stream char)
  (declare (ignore char))
  (let ((new-string "")
        (skip-if-quote nil))
    (loop (let ((char (read-char stream nil #\" t)))
            (if (char= char #\")
                (if (not skip-if-quote)
                    (return new-string)))
            (setf skip-if-quote nil)
            (if (char= char #\\)
                (setf skip-if-quote t))
            (setf new-string (concatenate 'string new-string (string char)))))))

(defun make-read-form-in-table-reader (to-readtable until-char &optional to-package)
  (lambda (stream char)
    (declare (ignore char))
    (let ((*readtable* to-readtable)
          (*package* (if to-package (find-package to-package) *package*)))
      (let ((result (read-delimited-list until-char stream t)))
	(assert (= (length result) 1))
	(car result)))))

(defun prepare-new-readtable ()
  (let ((c-readtable (copy-readtable))
        (normal-readtable (copy-readtable)))
    (setf (readtable-case c-readtable) :preserve)
    (set-macro-character #\" 'quote-reader nil c-readtable)
    (set-macro-character #\}
			 (lambda (stream char)
			   (declare (ignore stream char))
			   (error "Bare } found")) nil c-readtable)
    (set-macro-character #\]
			 (lambda (stream char)
			   (declare (ignore stream char))
			   (error "Bare ] found")) nil normal-readtable)
    (set-macro-character #\[
			 (make-read-form-in-table-reader normal-readtable #\] :scexp-user) nil c-readtable)
    (set-macro-character #\{
			 (make-read-form-in-table-reader c-readtable #\} :scexp-read) nil normal-readtable)
    (values c-readtable normal-readtable)))

(defun install-scexp-read-macro ()
  (setf *readtable* (nth-value 1 (prepare-new-readtable))))

(defun scexp-read-from-string (str)
  (let ((*package* (find-package :scexp-read))
        (*readtable* (prepare-new-readtable)))
    (read-from-string str)))

(defun translate-typespec (spec name &key (register-allowed t) (qualifiers nil))
  (labels ((output-qualifiers (rqual)
             (format nil "~{~A ~}"
                     (mapcar #'(lambda (e)
                                 (cdr (assoc e '((const . "const")
                                                 (pointer . "*")
                                                 (volatile . "volatile")
                                                 (unsigned . "unsigned")
                                                 (signed . "signed")
                                                 (long . "long")
                                                 (short . "short"))))) rqual))))
    (cond
      ((consp spec)
       (case (car spec)
         (scexp-read::|register|
                      (if register-allowed
                          (concatenate 'string "register "
                                       (translate-typespec (cdr spec) name :register-allowed nil))
                          (error "register not allowed here")))
         (scexp-read::|const|
                      (translate-typespec (cdr spec) name :register-allowed nil
                                          :qualifiers (cons 'const qualifiers)))
         (scexp-read::|pointer|
                      (translate-typespec (cdr spec) name :register-allowed nil
                                          :qualifiers (cons 'pointer qualifiers)))
         (scexp-read::|*| (translate-typespec (cdr spec) name :register-allowed nil
                                              :qualifiers (cons 'pointer qualifiers)))
         (scexp-read::|volatile| (translate-typespec (cdr spec) name :register-allowed nil
                                                     :qualifiers (cons 'volatile qualifiers)))
         (scexp-read::|signed| (translate-typespec (cdr spec) name :register-allowed nil
                                                     :qualifiers (cons 'signed qualifiers)))
         (scexp-read::|unsigned| (translate-typespec (cdr spec) name :register-allowed nil
                                                     :qualifiers (cons 'unsigned qualifiers)))
         (scexp-read::|long| (translate-typespec (cdr spec) name :register-allowed nil
                                                     :qualifiers (cons 'long qualifiers)))
         (scexp-read::|short| (translate-typespec (cdr spec) name :register-allowed nil
                                                     :qualifiers (cons 'short qualifiers)))
         (scexp-read::|function|
                      (let ((rqual (nreverse qualifiers)))
                        (destructuring-bind (out &rest in) (cdr spec)
                          (concatenate 'string
                                       (translate-typespec out nil :register-allowed nil)
                                       " ("
                                       (output-qualifiers rqual)
                                       (format nil "~@[~A~]) (~{~A~^, ~})" (if name (munge-symbol-name (symbol-name name)))
                                               (mapcar #'(lambda (e)
                                                           (translate-typespec e nil :register-allowed nil)) in))))))
         (scexp-read::|struct|
                      (let ((rqual (nreverse qualifiers)))
                        (destructuring-bind (structname) (cdr spec)
                          (format nil "struct ~A ~A~@[ ~A~]"
                                  (munge-symbol-name (symbol-name structname))
				  (output-qualifiers rqual)
                                  (if name (munge-symbol-name (symbol-name name)))))))
         (otherwise (if (null (cdr spec))
                        (translate-typespec (car spec) name :qualifiers qualifiers)
                        (error "Found unknown type: ~A" spec)))))
      (t (format nil "~A ~A~@[ ~A~]" (munge-symbol-name (symbol-name spec)) (output-qualifiers qualifiers)
                 (if name (munge-symbol-name (symbol-name name))))))))

(defvar *macros* nil) ; will be alist macro->expander
  
(defgeneric translate-form (type form-body))
  
(defmethod translate-form ((type (eql 'scexp-read::|deftype|))
                           form-body)
  (destructuring-bind (to from) form-body
    (format nil "typedef ~A;~%"
            (translate-typespec from to))))

(defmethod translate-form ((type (eql 'scexp-read::|defun|))
                           form-body)
  (destructuring-bind (out name in &body implicit-let) form-body
    (format nil "~A ~A(~{~A~^,~}) ~A"
            (translate-typespec out nil)
            (munge-symbol-name (symbol-name name))
            (mapcar #'(lambda (e)
                        (destructuring-bind (type name) e
                          (translate-typespec type name))) in)
            (translate-expr `(scexp-read::|let| ,@implicit-let)))))

(defmethod translate-form ((type (eql 'scexp-read::|defvar|))
                           form-body)
  (destructuring-bind (type name &optional val) form-body
      (format nil "~A~@[ = ~A~];"
              (translate-typespec type name)
              (if val (translate-expr-returning-value val)))))

(defun translate-struct (vals &optional name)
  (format nil "struct ~@[~A ~]{~{~A;~}}"
          (if name (munge-symbol-name (symbol-name name)))
          (mapcar #'(lambda (e)
                      (destructuring-bind (type name) e
                        (translate-typespec type name))) vals)))

(defmethod translate-form ((type (eql 'scexp-read::|defstruct|))
                           form-body)
  (destructuring-bind (name &rest vals) form-body
      (concatenate 'string (translate-struct vals name) ";")))

(defmethod translate-form ((type (eql 'scexp-read::|deftype/struct|))
                           form-body)
  (destructuring-bind (name &rest vals) form-body
    (format nil "typedef ~A ~A;" (translate-struct vals) (munge-symbol-name (symbol-name name)))))

(defmethod translate-form ((type (eql 'scexp-read::|cpp-include|))
                           form-body)
  (if (eq (car form-body) :|system|)
      (format nil "#include <~A>" (cadr form-body))
      (format nil "#include ~S" (car form-body))))

(defmethod translate-form ((type (eql 'defmacro))
                           form-body)
  (destructuring-bind (name lambda-list &body body) form-body
    (let ((expander (compile nil `(lambda ,lambda-list ,@body))))
      (aif (assoc name *macros*)
           (setf (cdr it) expander)
           (push (cons name expander) *macros*))
      (format nil "/* defined macro ~A */~%" name))))

(defmethod translate-form (type form-body)
  (aif (assoc type *macros*)
       (translate-toplevel-expr (apply (cdr it) form-body))
       (format *trace-output* "Warning: no translation for form (~A ~A)!~%" type form-body)))

(defun translate-toplevel-expr (expr)
  (translate-form (car expr) (cdr expr)))

(defun munge-symbol-name (name)
  (substitute #\_ #\- name))

(defun translate-expr-returning-value (expr)
  (labels ((compose-binary (format-rule values)
             (reduce #'(lambda (a b)
                         (format nil format-rule
                                 a
                                 b)) (mapcar #'translate-expr-returning-value values))))
    (cond
      ((and (consp expr)
            (assoc (car expr) *macros*))
       (translate-expr-returning-value (apply (cdr (assoc (car expr) *macros*))
                                              (cdr expr))))
      ((consp expr)
       (case (car expr)
         (scexp-read::+ (if (not (null (cddr expr)))
                            (compose-binary "(~A + ~A)" (cdr expr))
                            (format nil "(+~A)" (translate-expr-returning-value (cadr expr)))))
         (scexp-read::- (if (not (null (cddr expr)))
                            (compose-binary "(~A - ~A)" (cdr expr))
                            (format nil "(-~A)" (translate-expr-returning-value (cadr expr)))))
         (scexp-read::* (if (not (null (cddr expr)))
                            (compose-binary "(~A * ~A)" (cdr expr))
                            (format nil "(*~A)" (translate-expr-returning-value (cadr expr)))))
         (scexp-read::|setq| (compose-binary "(~A = ~A)" (cdr expr)))
	 (scexp-read::|setf| (compose-binary "(~A = ~A)" (cdr expr)))
         (scexp-read::== (compose-binary "(~A == ~A)" (cdr expr)))
         (scexp-read::< (compose-binary "(~A < ~A)" (cdr expr)))
         (scexp-read::> (compose-binary "(~A > ~A)" (cdr expr)))
         (scexp-read::<= (compose-binary "(~A <= ~A)" (cdr expr)))
         (scexp-read::>= (compose-binary "(~A >= ~A)" (cdr expr)))
         (scexp-read::!= (compose-binary "(~A != ~A)" (cdr expr)))
         (scexp-read::|/| (compose-binary "(~A / ~A)" (cdr expr)))
         (scexp-read::|mod| (compose-binary "(~A % ~A)" (cdr expr)))
         (scexp-read::|not| (if (null (cddr expr)) (format nil "(!~A)"
                                                           (translate-expr-returning-value (cadr expr)))
                                (error "not is one-argument only")))
         (scexp-read::|deref| (if (null (cddr expr)) (format nil "(*~A)"
                                                             (translate-expr-returning-value (cadr expr)))
                                  (error "deref is one-argument only")))
         (scexp-read::|incf| (if (null (cddr expr)) (format nil "(++~A)"
                                                            (translate-expr-returning-value (cadr expr)))
                                 (error "incf is one-argument only")))
         (scexp-read::|post-incf| (if (null (cddr expr)) (format nil "(~A++)"
                                                                 (translate-expr-returning-value (cadr expr)))
                                      (error "incf is one-argument only")))
         (scexp-read::|decf| (if (null (cddr expr)) (format nil "(--~A)"
                                                            (translate-expr-returning-value (cadr expr)))
                                 (error "incf is one-argument only")))
         (scexp-read::|post-decf| (if (null (cddr expr)) (format nil "(~A--)"
                                                                 (translate-expr-returning-value (cadr expr)))
                                      (error "incf is one-argument only")))
         (scexp-read::|log-and| (compose-binary "(~A && ~A)" (cdr expr)))
         (scexp-read::|log-or| (compose-binary "(~A || ~A)" (cdr expr)))
         (scexp-read::|bit-and| (compose-binary "(~A & ~A)" (cdr expr)))
         (scexp-read::|bit-or| (compose-binary "(~A | ~A)" (cdr expr)))
         (scexp-read::|bit-xor| (compose-binary "(~A ^ ~A)" (cdr expr)))
         (scexp-read::|if| (apply #'format nil "(~A ? ~A : ~A)"
                                  (mapcar #'translate-expr-returning-value (cdr expr))))
         (scexp-read::|progn| (if (cdr expr)
                                  (format nil "(~{~A~^, ~})"
                                          (mapcar #'translate-expr-returning-value (cdr expr)))
                                  ""))
         (scexp-read::|sizeof| (format nil "sizeof(~A)"
                                       (translate-typespec (cdr expr) nil :register-allowed nil)))
         (scexp-read::|aref|
                      (destructuring-bind (ptr &rest refs) (cdr expr)
                        (format nil "~A~{[~A]~}"
                                (translate-expr-returning-value ptr)
                                (mapcar
                                 #'translate-expr-returning-value refs))))
         (scexp-read::|ref|
                      (format nil "&~A" (translate-expr-returning-value (cadr expr))))
         (scexp-read::|cond|
                      (destructuring-bind ((test &body vals) &rest more) (cdr expr)
                        (if (eq test 'scexp-read::|else|)
                            (translate-expr-returning-value `(scexp-read::|progn| ,@vals))
                            (translate-expr-returning-value
                             `(scexp-read::|if| ,test (scexp-read::|progn| ,@vals)
                                           (scexp-read::|cond| ,@more))))))
         (scexp-read::|elt.|
                      (destructuring-bind (struct elt) (cdr expr)
                        (format nil "~A.~A"
                                (translate-expr-returning-value struct)
                                (translate-expr-returning-value elt))))
         (scexp-read::|elt->|
                      (destructuring-bind (struct elt) (cdr expr)
                        (format nil "~A->~A"
                                (translate-expr-returning-value struct)
                                (translate-expr-returning-value elt))))
         (scexp-read::|cast|
                      (destructuring-bind (val type) (cdr expr)
                        (format nil "(~A)~A"
                                (translate-typespec type nil)
                                (translate-expr-returning-value val))))
         (otherwise (format nil "~A(~{~A~^, ~})"
                            (symbol-name (car expr))
                            (mapcar #'translate-expr-returning-value (cdr expr))))))
      (t
       (if (symbolp expr)
           (munge-symbol-name (symbol-name expr))
           (if (stringp expr)
               (concatenate 'string "\"" (princ-to-string expr) "\"")
               (prin1-to-string expr)))))))

(defun translate-expr (expr)
  (cond
    ((and (consp expr)
          (assoc (car expr) *macros*))
     (translate-expr (apply (cdr (assoc (car expr) *macros*))
                            (cdr expr))))
    ((consp expr)
     (case (car expr)
       (scexp-read::|if|
                    (destructuring-bind (test then &optional else) (cdr expr)
                      (format nil "if (~A)~%~A~@[~%else ~A~]~%"
                              (translate-expr-returning-value test)
                              (translate-expr then)
                              (if else (translate-expr else)))))
       (scexp-read::|for| (destructuring-bind
                                (init test incs &body body) (cdr expr)
                            (format nil "for (~@[~A~];~@[~A~];~@[~A~])~%{~{~A~}}~%"
                                    (if init (translate-expr-returning-value `(scexp-read::|progn| ,@init)))
                                    (if test (translate-expr-returning-value test))
                                    (if incs (translate-expr-returning-value `(scexp-read::|progn| ,@incs)))
                                    (mapcar #'translate-expr body))))
       (scexp-read::|progn| (format nil "{~{~A~}}~%"
                                    (mapcar #'translate-expr (cdr expr))))
       (scexp-read::|do/while| (destructuring-bind
                                     ((&body body) test-form) (cdr expr)
                                 (format nil "do ~A~%while (~A);~%"
                                         (translate-expr `(scexp-read::|progn| ,@body))
                                         (translate-expr-returning-value test-form))))
       (scexp-read::|while| (destructuring-bind
                                  (test &body body) (cdr expr)
                              (format nil "while (~A)~%~A~%"
                                      (translate-expr-returning-value test)
                                      (translate-expr `(scexp-read::|progn| ,@body)))))
       (scexp-read::|let| (destructuring-bind
                                ((&rest vals) &body body) (cdr expr)
                            (format nil "{~{~A~}~{~A~}}~%"
                                    (mapcar #'(lambda (b)
                                                (destructuring-bind
                                                      (type name &optional binding) b
                                                  (format nil "~A~@[ = ~A~];~%"
                                                          (translate-typespec type name)
                                                          (if binding
                                                              (translate-expr-returning-value binding))))) vals)
                                    (mapcar #'translate-expr body))))
       (scexp-read::|cond|
                    (destructuring-bind ((test &body vals) &rest more) (cdr expr)
                      (if (eq test 'scexp-read::|else|)
                          (translate-expr `(scexp-read::|progn| ,@vals))
                          (translate-expr
                           `(scexp-read::|if| ,test (scexp-read::|progn| ,@vals)
                                         ,(if more
                                              `(scexp-read::|cond| ,@more)
                                              `(scexp-read::|progn|)))))))
       (scexp-read::|switch|
                    (destructuring-bind (on &body body) (cdr expr)
                      (format nil "switch (~A)~%{~{~A;~%~}}~%"
                              (translate-expr-returning-value on)
                              (mapcar #'translate-expr body))))
       (scexp-read::|case|
                    (destructuring-bind (what) (cdr expr)
                      (format nil "case ~A: "
                              (translate-expr-returning-value what))))
       (scexp-read::|break|
                    (format nil "break;~%"))
       (scexp-read::|default|
                    (format nil "default: "))
       (scexp-read::|label|
                    (destructuring-bind (what) (cdr expr)
                      (format nil "~A:" (munge-symbol-name (symbol-name what)))))
       (scexp-read::|goto|
                    (destructuring-bind (what) (cdr expr)
                      (format nil "goto ~A;~%" (munge-symbol-name (symbol-name what)))))
       (otherwise (concatenate 'string
                               (translate-expr-returning-value expr) ";" (string #\newline)))))
    (t (concatenate 'string
                    (translate-expr-returning-value expr) ";" (string #\newline)))))

(defun translate-file (file-name &key (output-file
                                       (make-pathname :type "c"
                                                      :defaults file-name)))
  (with-open-file (file (merge-pathnames file-name
                                         (make-pathname :type "scexp")) :direction :input)
    (with-open-file (output output-file :direction :output
                            :if-exists :supersede)
      (let ((*package* (find-package :scexp-read))
            (*readtable* (prepare-new-readtable))
            (*macros* nil)
            (eof (gensym)))
        (loop (let ((form (read file nil eof)))
                (if (not (eq form eof))
                    (progn
                      (princ (translate-toplevel-expr form) output)
                      (terpri output))
                    (return nil))))))))
